﻿<#
.SYNOPSIS
Script for migrating a VM from VMWare to Hyper-V using MVMC

.DESCRIPTION
This script will migrate a VM from VMWare to Hyper-V using MVMC. The script also installs the Hyper-V Integration Services, restores network configuration, and removes VMWare Tools.
#>
param
(
    #VIServer and credentials
    [Parameter(Mandatory=$true)]
    [hashtable]$viserver,

	#Destination VMHost
    [Parameter(Mandatory=$true)]
    [string]$VMHost,
	
    #Destination Location
    [Parameter(Mandatory=$true)]
    [string]$destLoc,

    #debug mode
    [switch]$VerboseMode
)

function Log($out) { 
    $out = [System.DateTime]::Now.ToString("yyyy.MM.dd hh:mm:ss") + " ---- " + $out; 
    Write-Output "$out"; 
}

function Log-Verbose($out) {
	if ($VerboseMode) {
		Log $out
	}
}

function Test-RegistryValue {
    <#
	.DESCRIPTION
	Tests for existence of registry value
	
	.PARAMETER Path
	[string] Registry Key

    .PARAMETER Value
    [string] Name of Registry Value
	
	.EXAMPLE
	Test-RegistryValue -Path 'HKLM:\SOFTWARE\TestSoftware' -Value 'Version'
	#>
    param (
        [parameter(Mandatory=$true)]
        [ValidateNotNullOrEmpty()]$Path,

        [parameter(Mandatory=$true)]
        [ValidateNotNullOrEmpty()]$Value
    )
    try {
        $c = 0
        Get-Item -Path $Path | Select-Object -ExpandProperty property | % { if ($_ -match $Value) { $c=1 ; return $true} }
        if ($c -eq 0) {
            return $false
        }
    }
    catch {
        return $false;
    }
}

Log-Verbose "Parameters";
Log-Verbose "VIServer:`t`t$($viserver.Server)";
Log-Verbose "Username:`t`t$($viserver.Credential.Username)";
Log-Verbose "Password:`t`t*******";
Log-Verbose "VMHost:`t`t$VMHost";
Log-Verbose "DestLoc:`t`t$destLoc";

$script:viconnection = $null;
$script:regsys = "VHDSYS";
$script:regsw = "VHDSW";
$script:helperfiles = "client-tools";

Log-Verbose "Guest Reg SYS:`t$($script:regsys)";
Log-Verbose "Guest Reg SW:`t$($script:regsw)";
Log-Verbose "HelperFiles:`t$($script:helperfiles)";

Log "Loading MVMC Module";
try {
    if(!(Test-Path "C:\Program Files\Microsoft Virtual Machine Converter\MvmcCmdlet.psd1")) {
        throw "MVMC version 2.0+ required.";
    }
    Import-Module "C:\Program Files\Microsoft Virtual Machine Converter\MvmcCmdlet.psd1" -ErrorAction Stop;
    $module = Get-Module -Name MvmcCmdlet;
    if ($module.Version -notmatch "2\..*") {
        throw "MVMC version 2.0+ required. Found version: $($module.version)";
    }
}
catch {
    Log "Error loading MVMC: $($error[0])";
}

Log "Loading Hyper-V Module";
try {
    Import-Module "Hyper-V" -ErrorAction Stop;
    $module = Get-Module -Name Hyper-V;
    if ($module.Version -notmatch "1\.1") {
        throw "Hyper-V Powershell cmdlets version 1.1 required. Found version: $($module.version)";
    }
}
catch {
    Log "Error loading Hyper-V cmdlets: $($error[0])";
}

Log "Loading PowerCLI library";
try {
    Add-PSSnapin -Name VMware.VimAutomation.Core -ErrorAction Stop;
    $snapin = Get-PSSnapin -Name VMware.VimAutomation.Core;
    if ($snapin.Version -notmatch "5\.8\..*") {
        throw "VMware.VimAutomation.Core version 5.8 required. Found version: $($snapin.version)";
    }
}
catch {
    Log "Error loading PowerCLI: $($error[0])";
}

Log "Connecting to VIServer: $($viserver.Server)";
try {
    $script:viconnection = VMware.VimAutomation.Core\Connect-VIServer -Server $viserver.Server -Credential $viserver.Credential -NotDefault -ErrorAction Stop;
}
catch {
    Log "Error loading PowerCLI: $($error[0])";
}

function Get-VMDetails {
	<#
	.DESCRIPTION
	Retrieves detailed VM configuration from VMWare VM
	
	.PARAMETER sourcevm
	VMware.VimAutomation.ViCore.Impl.V1.Inventory.VirtualMachineImpl
	
	.EXAMPLE
	$vm = VMware.VimAutomation.Core\Get-VM MyVM
	Get-VMDetails $vm
	#>
	param
	(
		#VMware.VimAutomation.ViCore.Impl.V1.Inventory.VirtualMachineImpl
		[Parameter(Mandatory=$true)]
		$sourcevm
	)
    $disks = Get-VMDisks $sourcevm;

    $nics = Get-VMNetworkAdapters $sourcevm;

    $vm = New-Object -TypeName PSObject;
    $vm | Add-Member -MemberType NoteProperty -Name Name -Value $sourcevm.Name;
    $vm | Add-Member -MemberType NoteProperty -Name vCPU -Value $sourcevm.NumCpu;
    $vm | Add-Member -MemberType NoteProperty -Name vRAM -Value $sourcevm.MemoryMB;
    $vm | Add-Member -MemberType NoteProperty -Name GuestOS -Value $sourcevm.Guest.OSFullName;
    $vm | Add-Member -MemberType NoteProperty -Name GuestId -Value $sourcevm.GuestId
    $vm | Add-Member -MemberType NoteProperty -Name Disks -Value $disks;
    $vm | Add-Member -MemberType NoteProperty -Name Nics -Value $nics;
	
	return $vm;
}

function Get-VMNetworkAdapters {
	<#
	.DESCRIPTION
	Retrieves basic network adapter infomration from VMWare VM and returns custom PSObject
	
	.PARAMETER vm
	VMware.VimAutomation.ViCore.Impl.V1.Inventory.VirtualMachineImpl
	
	.EXAMPLE
	$vm = VMware.VimAutomation.Core\Get-VM MyVM
	Get-VMNetworkAdapters $vm
	#>
	param
	(
		#VMware.VimAutomation.ViCore.Impl.V1.Inventory.VirtualMachineImpl
		[Parameter(Mandatory=$true)]
		$vm
	)
	
	$nics = @();
    for($i=0;$i -lt $vm.NetworkAdapters.Count;$i++) {
        $adapter = $vm.NetworkAdapters[$i];
        $nic = New-Object -TypeName PSObject;
        $nic | Add-Member -MemberType NoteProperty -Name Name -Value $adapter.Name;
        $nic | Add-Member -MemberType NoteProperty -Name Network -Value $adapter.NetworkName;
        $nic | Add-Member -MemberType NoteProperty -Name MacAddress -Value $adapter.MacAddress;
        $nic | Add-Member -MemberType NoteProperty -Name NicType -Value $adapter.Type;

        #SlotInfo is not populated on VM's with a single nic
        $slot = @{$true=$adapter.ExtensionData.SlotInfo.PciSlotNumber;$false=($adapter.Parent.ExtensionData.Config.ExtraConfig | ?{$_.Key -match "ethernet$($i).pciSlotNumber"}).Value}[$adapter.ExtensionData.SlotInfo.PciSlotNumber -ne $null];
        $nic | Add-Member -MemberType NoteProperty -Name PciSlotNumber -Value $slot;
		
        #VlanId is set at the port group level
        $vlanid = (Get-VirtualPortGroup -Server $script:viconnection | ?{$_.Name -eq "$($adapter.NetworkName)"}).VlanId
		$nic | Add-Member -MemberType NoteProperty -Name VlanId -Value $vlanid;
		
        $nics += $nic;
    }
	return $nics;
}

function Get-VMNetworkAdaptersDetail {
	<#
	.DESCRIPTION
	Populates detailed nic information from Windows registry for all adapters and returns custom PSObject array
	
	.PARAMETER adapters
	PSCustomObject Array of network adapters
	#>
	param
	(
		#PSCustomObject Array of network adapters
		[Parameter(Mandatory=$true)]
		$adapters
	)
    $nics = @();
    foreach ($adapter in $adapters) {
        $nicinfo = Get-NICDeviceInfo-FromRegistry $($adapter.PciSlotNumber);
        if ($nicinfo) {
            $adapter | Add-Member -MemberType NoteProperty -Name UINumber -Value $nicinfo.UINumber;
            $adapter | Add-Member -MemberType NoteProperty -Name DeviceID -Value $nicinfo.DeviceID;
            $adapter | Add-Member -MemberType NoteProperty -Name DHCP -Value $nicinfo.DHCP;
            $adapter | Add-Member -MemberType NoteProperty -Name IPAddress -Value $nicinfo.IPAddress;
            $adapter | Add-Member -MemberType NoteProperty -Name SubnetMask -Value $nicinfo.SubnetMask;
            $adapter | Add-Member -MemberType NoteProperty -Name Gateway -Value $nicinfo.DefaultGateway;
            $adapter | Add-Member -MemberType NoteProperty -Name GatewayMetric -Value $nicinfo.DefaultGatewayMetric;
            $adapter | Add-Member -MemberType NoteProperty -Name NameServer -Value $nicinfo.NameServer;
        }

        $nics += $adapter;
    }
    return $nics;
}

function Get-UINumber {
	<#
	.DESCRIPTION
	Converts to VMWare PciSlotNumber to UINumber
	
	.PARAMETER slot
	[int] The PCI Slot number from VMware.Vim.VirtualDevicePciBusSlotInfo
		
	.NOTES
	http://kb.vmware.com/selfservice/microsites/search.do?language=en_US&cmd=displayKC&externalId=2047927
	#>
	param
	(
		#PciSlotNumber
		[Parameter(Mandatory=$true)]
		[int] $slot
	)
	$x = $slot -band 7168;
	$base = $slot -bxor $x;
	return $($base + $($x / 1024));
}

function Get-NICDeviceInfo-FromRegistry {
	<#
	.DESCRIPTION
	Retrieves detailed nic information from Windows registry and returns custom PSObject
	
	.PARAMETER slot
	[int] The PCI Slot number from VMware.Vim.VirtualDevicePciBusSlotInfo
		
	.NOTES
	Function assumes VHDX has been mounted, and guest registry loaded to HKLM:\$($script:regsys)
	#>
	param
	(
		#PciSlotNumber
		[Parameter(Mandatory=$true)]
		[int] $slot
	)
	$nicinfo = $null
	$uinum = Get-UINumber $slot;
	$devices = Get-ChildItem "HKLM:\$($script:regsys)\ControlSet001\Enum\PCI" -ErrorAction SilentlyContinue;
    $interfaces = Get-ChildItem "HKLM:\VHDSYS\ControlSet001\Control\Class\{4d36e972-e325-11ce-bfc1-08002be10318}" -ErrorAction SilentlyContinue;
	
    foreach($device in $devices) {
		$instances = @(Get-ChildItem "$($device.PSPath)" -ErrorAction SilentlyContinue | ? {$_.GetValue("UINumber") -eq $uinum;})
		if ($instances.count -eq 1) {
			$instance = $instances[0];
			$deviceid = $instance.PSPath -replace "Microsoft.PowerShell.Core\\Registry::HKEY_LOCAL_MACHINE\\$($script:regsys)\\ControlSet001\\Enum\\", "";
						
			$nicinfo = New-Object -TypeName PSObject;
			$nicinfo | Add-Member -MemberType NoteProperty -Name Slot -Value $slot;
			$nicinfo | Add-Member -MemberType NoteProperty -Name UINumber -Value $uinum;
			$nicinfo | Add-Member -MemberType NoteProperty -Name DeviceID -Value $deviceid;
			
			$nics = @($interfaces | ? {$_.GetValue("DeviceInstanceId") -eq $deviceid});
			if ($nics.count -eq 1) {
				$nic = $nics[0];
				$netcfginstanceid = $nic.GetValue("NetCfgInstanceId");
				$nicinfo | Add-Member -MemberType NoteProperty -Name NetCfgInstanceId -Value $netcfginstanceid;
				
				$cfgpath = (Get-Item "HKLM:\$($script:regsys)\ControlSet001\Services\Tcpip\Parameters\Adapters\$netcfginstanceid").GetValue("IpConfig");
				if ($cfgpath) {
					$ipconfig = (Get-Item "HKLM:\$($script:regsys)\ControlSet001\Services\$cfgpath");
					$nicinfo | Add-Member -MemberType NoteProperty -Name DHCP -Value ($ipconfig.GetValue("EnableDHCP") -eq $true);
					$nicinfo | Add-Member -MemberType NoteProperty -Name IPAddress -Value @{$true=@($ipconfig.GetValue("IPAddress"));$false=$null}[$ipconfig.GetValue("IPAddress").Count -gt 0];
					$nicinfo | Add-Member -MemberType NoteProperty -Name SubnetMask -Value @{$true=@($ipconfig.GetValue("SubnetMask"));$false=$null}[$ipconfig.GetValue("SubnetMask").Count -gt 0];
					$nicinfo | Add-Member -MemberType NoteProperty -Name DefaultGateway -Value @{$true=@($ipconfig.GetValue("DefaultGateway"));$false=$null}[$ipconfig.GetValue("DefaultGateway").Count -gt 0];
					$nicinfo | Add-Member -MemberType NoteProperty -Name DefaultGatewayMetric -Value @{$true=@(($ipconfig.GetValue("DefaultGatewayMetric") | %{@{$true=1;$false=$_}[$_ -eq 0]}));$false=$null}[$ipconfig.GetValue("DefaultGatewayMetric").Count -gt 0];
					$nicinfo | Add-Member -MemberType NoteProperty -Name NameServer -Value @($ipconfig.GetValue("NameServer").Split(","));
				}
			}
			break;
		}
	}

    #Unloading a registry hive may fail if we have open references to registry objects
    Remove-Variable devices -ErrorAction SilentlyContinue;
    Remove-Variable device -ErrorAction SilentlyContinue;
    Remove-Variable interfaces -ErrorAction SilentlyContinue;
    Remove-Variable instances -ErrorAction SilentlyContinue;
    Remove-Variable instance -ErrorAction SilentlyContinue;
    Remove-Variable nics -ErrorAction SilentlyContinue;
    Remove-Variable nic -ErrorAction SilentlyContinue;
    Remove-Variable netcfginstanceid -ErrorAction SilentlyContinue;
    Remove-Variable cfgpath -ErrorAction SilentlyContinue;
    Remove-Variable ipconfig -ErrorAction SilentlyContinue;

    if($nicinfo -eq $null) {
        $nicinfo = New-Object -TypeName PSObject;
    }

    if($nicinfo.DHCP -eq $null) {
        $nicinfo | Add-Member -MemberType NoteProperty -Name DHCP -Value $true;
    }

	return $nicinfo;
}

function Get-VMDisks {
	<#
	.DESCRIPTION
	Retrieves hard disk information from VMWare VM and returns custom PSObject
	
	.PARAMETER vm
	VMware.VimAutomation.ViCore.Impl.V1.Inventory.VirtualMachineImpl
	
	.EXAMPLE
	$vm = VMware.VimAutomation.Core\Get-VM MyVM
	Get-VMDisks $vm
	
	.NOTES
	Attempts to determine boot disk from Config.BootOptions data. If that data is not available, the function will return the first available boot disk
	#>
	param
	(
		#VMware.VimAutomation.ViCore.Impl.V1.Inventory.VirtualMachineImpl
		[Parameter(Mandatory=$true)]
		$vm
	)
	
    $disks = @();
    foreach ($vdisk in $vm.HardDisks) {
        $disk = New-Object -TypeName PSObject;
        $disk | Add-Member -MemberType NoteProperty -Name Name -Value $vdisk.Name;
        $disk | Add-Member -MemberType NoteProperty -Name DeviceKey -Value $vdisk.ExtensionData.Key;
        $disk | Add-Member -MemberType NoteProperty -Name ControllerKey -Value $vdisk.ExtensionData.ControllerKey;
        $disk | Add-Member -MemberType NoteProperty -Name Capacity -Value $vdisk.CapacityGB;

        $diskloc = Get-VMDiskBusLocation $vdisk;

        if(!$diskloc) {
            $diskloc = @{Bus="none";BusNum="99";UnitNumber="256"};
        }

        $disk | Add-Member -MemberType NoteProperty -Name Bus -Value $diskloc.Bus;
        $disk | Add-Member -MemberType NoteProperty -Name BusNum -Value $diskloc.BusNum;
        $disk | Add-Member -MemberType NoteProperty -Name UnitNumber -Value $diskloc.UnitNumber;

        $disk | Add-Member -MemberType NoteProperty -Name BootDevice -Value $false;

        $disks += $disk;
    }

    if($vm.ExtensionData.Config.BootOptions.BootOrder.Count -gt 0) {
        $bootdisks = @($vm.ExtensionData.Config.BootOptions.BootOrder | ?{($disks | Select DeviceKey).DeviceKey -Contains $_.DeviceKey});
        if($bootdisks.Count -gt 0) {
            $bootdisk = ($vm.ExtensionData.Config.BootOptions.BootOrder | ?{($disks | Select DeviceKey).DeviceKey -Contains $_.DeviceKey})[0].DeviceKey;
        }
    }

    if (!$bootdisk) {
        $bootdisk = ($disks | Sort Bus, BusNum, UnitNumber)[0].DeviceKey;
    }

    $disks | ?{$_.DeviceKey -eq $bootdisk } | % {$_.BootDevice = $true;}

    return $disks;
}

function Get-VMDiskBusLocation {
	<#
	.DESCRIPTION
	Retrieves bus location from VMWare VM Hard Disk and returns hashtable
	
	.PARAMETER disk
	VMware.VimAutomation.ViCore.Impl.V1.VirtualDevice.HardDiskImpl
	
	.EXAMPLE
	$vm = VMware.VimAutomation.Core\Get-VM MyVM
	Get-VMDiskBusLocation $vm.HardDisk[0]
	
	#>
	param
	(
		#VMware.VimAutomation.ViCore.Impl.V1.VirtualDevice.HardDiskImpl
		[Parameter(Mandatory=$true)]
		$disk
	)
	
    $idectrls = $disk.Parent.ExtensionData.Config.Hardware.Device | ?{$_.GetType().Name -eq "VirtualIDEController"; }
    $scsictrls = $disk.Parent.ExtensionData.Config.Hardware.Device | ?{$_.GetType().BaseType.Name -eq "VirtualSCSIController"; }

    $devicekey = $disk.ExtensionData.Key;

    $diskloc = $null;

    if($idectrls) {
        foreach($ctrl in $idectrls) {
            if($ctrl.Device.Count -gt 0) {
                if($ctrl.Device.IndexOf($devicekey) -ge 0) {
                    return @{Bus="ide";BusNum=$ctrl.BusNumber;UnitNumber=$ctrl.Device.IndexOf($devicekey)};
                    break;
                }
            }
        }
    }

    if($scsictrls) {
        foreach($ctrl in $scsictrls) {
            if($ctrl.Device.IndexOf($devicekey) -ge 0) {
                return @{Bus="scsi";BusNum=$ctrl.BusNumber;UnitNumber=$ctrl.Device.IndexOf($devicekey)};
                break;
            }
        }
    }

    return $diskloc;
}

function Migrate-VMWareVM {
	<#
	.DESCRIPTION
	Migrates a VM from VMWare to Hyper-V
	
	.PARAMETER vm
	PSCustomObject Get-VMDetails result
	#>
	param
	(
		#PSCustomObject Get-VMDetails
		[Parameter(Mandatory=$true)]
		$vm
	)
    Log "Migrating VMware VM $($vm.Name)";
    Log-Verbose "Getting connection to VMware VIServer: $($viserver.Server)";
    $mvmcconn = New-MvmcSourceConnection -Server $viserver.Server -SourceCredential $viserver.Credential;
    Log-Verbose "Getting VMware VM: $($vm.Name)";
    $sourcevm = Get-MvmcSourceVirtualMachine -SourceConnection $mvmcconn | ?{$_.Name -match $($vm.Name)};
	$IsWindowsOS = $sourcevm.IsWindowsOS;
    Log-Verbose "VmId:`t`t`t$($sourcevm.GuestVmId)";
    Log-Verbose "WindowsOS:`t`t$IsWindowsOS";
    Log-Verbose "HVSupportedOS:`t$($sourcevm.IsHyperVSupportedOS)";
    Log-Verbose "OS:`t`t`t$($sourcevm.OperatingSystem)";
    Log-Verbose "State:`t`t`t$($sourcevm.State)";

    Log-Verbose "Exporting $($vm.Name) to OVF: $destLoc";
    $convertedvm = ConvertTo-MvmcVirtualHardDiskOvf -SourceConnection $mvmcconn -DestinationLiteralPath "$destLoc" -GuestVmId $sourcevm.GuestVmId -VhdType DynamicHardDisk -VhdFormat Vhdx -SourceVMPowerOption PowerOn;
	
	#We've likely disconnected at this point, need to reconnect
	$mvmcconn = New-MvmcSourceConnection -Server $viserver.Server -SourceCredential $viserver.Credential;
    $sourcevm = Get-MvmcSourceVirtualMachine -SourceConnection $mvmcconn | ?{$_.Name -match $($vm.Name)};
	
    Log-Verbose "Shutting down source VM $($vm.Name)";
    try {
	    Stop-MvmcSourceVirtualMachine -SourceConnection $mvmcconn -GuestVmId $sourcevm.GuestVmId -SourceVMPowerOption Shutdown -ErrorAction SilentlyContinue;
    }
    catch {
        Log-Verbose "Error shutting down VM: $($error[0].Exception.InnerException)";
    }

    Log-Verbose "Parsing OVF file to find bootdisk";
	$xml = [xml] (Get-Content "$($destLoc)\$($vm.Name)\mvmc.ovf");

    $resource = ($xml.Envelope.VirtualSystem.VirtualHardwareSection.Item | ?{$_.ElementName -match ($vm.Disks | ?{$_.BootDevice -eq $true}).Name;}).HostResource.Split("/")[2];
    $fileref = ($xml.Envelope.DiskSection.Disk | ?{$_.diskId -match $resource}).fileRef;
    $diskref = ($xml.Envelope.References.File | ?{$_.id -match $fileref}).href;

    $bootdisk = "$($destLoc)\$($vm.Name)\$diskref";
    
    Log-Verbose "Preparing bootdisk: $bootdisk";
    Prepare-VHD -vm $vm -diskpath $bootdisk -IsWindowsOS $IsWindowsOS;
	
    Log-Verbose "Creating Hyper-V VM $($vm.Name)";
    Log-Verbose "VMPath:`t`t$destLoc";
    Log-Verbose "vCPUs:`t`t`t$($vm.vCPU)";
    Log-Verbose "vRAM:`t`t`t$($vm.vRAM) MB";
    Log-Verbose "Disks:`t`t`t$(@($vm.Disks).Count)";
    Log-Verbose "Nics:`t`t`t$(@($vm.Nics).Count)";
	$destvm = Hyper-V\New-VM -ComputerName $VMHost -Name $vm.Name -Generation 1 -Path $destLoc -MemoryStartupBytes $($vm.vRAM*1MB) -VHDPath $bootdisk -BootDevice VHD;
	$destvm | Hyper-V\Set-VMProcessor -Count $vm.vCPU;
	$destvm | Hyper-V\Remove-VMNetworkAdapter;
	
	foreach($disk in $vm.Disks) {
		$resource = ($xml.Envelope.VirtualSystem.VirtualHardwareSection.Item | ?{$_.ElementName -match $disk.Name;}).HostResource.Split("/")[2];
		$fileref = ($xml.Envelope.DiskSection.Disk | ?{$_.diskId -match $resource}).fileRef;
		$diskref = ($xml.Envelope.References.File | ?{$_.id -match $fileref}).href;

		$diskpath = "$($destLoc)\$($vm.Name)\$diskref";
		
		if($diskpath -ne $bootdisk) {
            Log-Verbose "Attaching disk: $diskpath";
			$destvm | Hyper-V\Add-VMHardDiskDrive -ControllerType SCSI -Path $diskpath;
		}
	}
	
	foreach($nic in $vm.Nics) {
        Log-Verbose "Attaching nic: $($nic.Name) ($($nic.MacAddress)) - Switch: $($nic.Network)";
		$adapter = $destvm | Hyper-V\Add-VMNetworkAdapter -StaticMacAddress $nic.MacAddress -SwitchName $nic.Network;
		if($nic.VlanId -ne 0) {
            Log-Verbose "Setting VLAN on on $($nic.Name): ($($nic.VlanId))";
			$adapter | Hyper-V\Set-VMNetworkAdapterVlan -Access -VlanId $nic.VlanId;
		}
	}

    Log "Starting migrated VM: $($vm.Name)";
    $destvm | Hyper-V\Start-VM;
}

function Prepare-VHD {
	<#
	.DESCRIPTION
	Mounts the guest VHDX, installs HV Integration Services, re-configures networking, removes VMWare Tools
	
	.PARAMETER vm
	PSCustomObject Get-VMDetails result
	
	.PARAMETER diskpath
	[string] Path to system VHDX
	
	.PARAMETER IsWindowsOS
	[bool] True for Windows OS, False otherwise	
	#>
	param
	(
		#PSCustomObject Get-VMDetails
		[Parameter(Mandatory=$true)]
		$vm,
		
		#Boolean, True for Windows OS, False otherwise
		[Parameter(Mandatory=$true)]
		[string] $diskpath,
		
		#Boolean, True for Windows OS, False otherwise
		[Parameter(Mandatory=$true)]
		[bool] $IsWindowsOS
	)

    if($IsWindowsOS) {
        Log-Verbose "Preparing Windows VHDX";

        $bitness = @{$true="x64";$false="x86"}[$vm.GuestId -match "64"];
        Log-Verbose "OS Bitness:`t$bitness";

        Log-Verbose "Mounting VHDX with Read/Write access: $diskpath";
        Mount-DiskImage -Access ReadWrite -ImagePath "$diskpath" -StorageType Vhdx;

        Log-Verbose "Searching for Windows installation";
        $drives = @(Get-DiskImage "$diskpath" | Get-Disk | Get-Partition | Get-Volume | ? FileSystemLabel -notmatch "System Reserved");

        foreach($drive in $drives) {
            if (Test-Path "$($drive.DriveLetter):\Windows") { 
                $windrive = $drive;
                break;
            }
        }
        
        if($windrive) {
            Log-Verbose "Found Windows installation on $($windrive.DriveLetter):\";
            
            Log-Verbose "Loading guest SYSTEM registry hive to HKLM\$($script:regsys)";
            reg load HKLM\$($script:regsys) "$($windrive.DriveLetter):\windows\system32\config\system" | Out-Null;
            if (!(Test-Path "HKLM:\$($script:regsys)")) {
                Log-Verbose "Failed to mount guest SYSTEM hive to HKLM\$($script:regsys)";
            }
            
            Log-Verbose "Loading guest SOFTWARE registry hive to HKLM\$($script:regsw)";
            reg load HKLM\$($script:regsw) "$($windrive.DriveLetter):\windows\system32\config\software" | Out-Null;
            if (!(Test-Path "HKLM:\$($script:regsw)")) {
                Log-Verbose "Failed to mount guest SOFTWARE registry hive to HKLM\$($script:regsw)";
            }

            if ((Test-Path "HKLM:\$($script:regsw)") -and (Test-Path "HKLM:\$($script:regsys)")) {
                Log-Verbose "Building helper script";
                $command = "";
                
                Log-Verbose "Adding helper command to install Hyper-V Integration Services";
                $command += "start /wait C:\Windows\Temp\$($script:helperfiles)\IS\support\$(@{$true="amd64";$false="x86"}[$bitness -eq "x64"])\setup.exe /quiet /norestart`r`n";

                Log-Verbose "Gathering guest nic device and IP information from guest registry";
                $nics = Get-VMNetworkAdaptersDetail $($vm.Nics);
                foreach($nic in $nics) {
                    if($nic.DeviceID -ne $null) {
                        Log-Verbose "Adding helper command to remove VMWare nic: $($nic.Name) ($($nic.MacAddress))";
                        $command += "start /wait C:\Windows\Temp\$($script:helperfiles)\devcon\$(@{$true="x64";$false="x86"}[$bitness -eq "x64"])\devcon.exe -r remove `"@$($nic.DeviceID)`"`r`n";
                    }
                    if(!$nic.DHCP) {
                        Log-Verbose "Adding helper command to restore IP settings to new Hyper-V nic: $($nic.Name) ($($nic.MacAddress))";
                        $command += "cscript C:\Windows\Temp\$($script:helperfiles)\client-network.vbs `"$($nic.MacAddress)`" `"$([string]::Join(",",$nic.IPAddress))`" `"$([string]::Join(",",$nic.SubnetMask))`" `"$([string]::Join(",",$nic.Gateway))`" `"$([string]::Join(",",$nic.GatewayMetric))`" `"$([string]::Join(",",$nic.NameServer))`"`r`n";
                    }
                }

                Log-Verbose "Searching guest registry for VMware Tools installtion";
                $progs = Get-ChildItem HKLM:\VHDSW\Microsoft\Windows\CurrentVersion\Uninstall
                $progs | %{ if ($_.GetValue("DisplayName") -match "VMware Tools") { $prog = $_; }}

                if($prog) {
                    Log-Verbose "Adding helper command to gracefully remove VMware Tools installation";
                    $command += "start /wait msiexec.exe /x$($prog.PSChildName) /qn /norestart`r`n";
                }
                else {
                    Log-Verbose "Did not find VMware Tools uninstall command in registry";
                    Log-Verbose "Adding helper command remove VMware Tools installation";
                    $command += "start /wait C:\Windows\Temp\$($script:helperfiles)\VMTools\setup.exe /c /s /v/qn`r`n";   
                }

                Log-Verbose "Adding helper command to cleanup";
                $command += "sc delete vmware-to-hyperv-helper`r`n";
                Log-Verbose "Adding helper command to reboot";
                $command += "shutdown /r /t 0`r`n";
                
                Log-Verbose "Writing helper script: $($windrive.DriveLetter):\Windows\Temp\vmware-to-hyperv-helper.cmd";
                Set-Content "$($windrive.DriveLetter):\Windows\Temp\vmware-to-hyperv-helper.cmd" $command -Encoding ASCII;

                $result = New-Item -Path "HKLM:\$($script:regsys)\ControlSet001\Services" -Name "vmware-to-hyperv-helper";
                $result = New-Item -Path "HKLM:\$($script:regsys)\ControlSet001\Services\vmware-to-hyperv-helper" -Name "Parameters";
                $result = New-ItemProperty "HKLM:\$($script:regsys)\ControlSet001\Services\vmware-to-hyperv-helper" -Name "Type" -PropertyType DWord -Value 16;
                $result = New-ItemProperty "HKLM:\$($script:regsys)\ControlSet001\Services\vmware-to-hyperv-helper" -Name "Start" -PropertyType DWord -Value 2;
                $result = New-ItemProperty "HKLM:\$($script:regsys)\ControlSet001\Services\vmware-to-hyperv-helper" -Name "ErrorControl" -PropertyType DWord -Value 1;
                $result = New-ItemProperty "HKLM:\$($script:regsys)\ControlSet001\Services\vmware-to-hyperv-helper" -Name "ImagePath" -PropertyType String -Value "C:\Windows\Temp\client-tools\nssm\$(@{$true="win64";$false="win32"}[$bitness -eq "x64"])\nssm.exe";
                $result = New-ItemProperty "HKLM:\$($script:regsys)\ControlSet001\Services\vmware-to-hyperv-helper" -Name "DisplayName" -PropertyType String -Value "VMWare to Hyper-V Conversion Helper";
                $result = New-ItemProperty "HKLM:\$($script:regsys)\ControlSet001\Services\vmware-to-hyperv-helper" -Name "ObjectName" -PropertyType String -Value "LocalSystem";
                $result = New-ItemProperty "HKLM:\$($script:regsys)\ControlSet001\Services\vmware-to-hyperv-helper\Parameters" -Name "AppDirectory" -PropertyType String -Value "C:\Windows\Temp";
                $result = New-ItemProperty "HKLM:\$($script:regsys)\ControlSet001\Services\vmware-to-hyperv-helper\Parameters" -Name "Application" -PropertyType String -Value "C:\Windows\System32\cmd.exe";
                $result = New-ItemProperty "HKLM:\$($script:regsys)\ControlSet001\Services\vmware-to-hyperv-helper\Parameters" -Name "AppParameters" -PropertyType String -Value "/c C:\Windows\Temp\vmware-to-hyperv-helper.cmd";

                if(Test-RegistryValue -Path "HKLM:\$($script:regsw)\Microsoft\Windows\CurrentVersion\RunOnce" -Value "InstallIntegrationServices") {
                    Log-Verbose "Clearing MVMC IS installation regkey";
                    $result = Remove-ItemProperty -Path "HKLM:\$($script:regsw)\Microsoft\Windows\CurrentVersion\RunOnce" -Name "InstallIntegrationServices";
                    $result = Remove-Item -Path "$($windrive.DriveLetter):\windows\system32\InstallIntegrationServices.bat";
                }
            }

            #Unloading a hive may fail if we don't run a garbage collection first
            Remove-Variable result -ErrorAction SilentlyContinue;
            Remove-Variable progs -ErrorAction SilentlyContinue;
            Remove-Variable prog -ErrorAction SilentlyContinue;
            [gc]::Collect();

            Log-Verbose "Unloading guest SYSTEM registry hive";
            reg unload HKLM\$($script:regsys) | Out-Null;

            Log-Verbose "Unloading guest SOFTWARE registry hive";
            reg unload HKLM\$($script:regsw) | Out-Null;

            Log-Verbose "Copying helper files from .\$($script:helperfiles) to $($windrive.DriveLetter):\Windows\Temp";
            Copy-Item -Recurse ".\$($script:helperfiles)" "$($windrive.DriveLetter):\Windows\Temp";
        }
        else {
            Log-Verbose "Could not find any Windows installations on VHDX: $diskpath";
        }

        Log-Verbose "Dismounting VHDX: $diskpath";
        Dismount-DiskImage "$diskpath";
    }
}
